@@ -75,11 +75,11 @@ data Config = Config
     , configMaxDiff         :: !Int
     , configCacheRefresh    :: !Int
     , configCacheRetryDelay :: !Int
-    , configNumTxId         :: !Bool
     , configStatsd          :: !Bool
     , configStatsdHost      :: !String
     , configStatsdPort      :: !Int
     , configStatsdPrefix    :: !String
+    , configWebPriceGet     :: !Int
     }
 
 instance Default Config where
@@ -107,11 +107,11 @@ instance Default Config where
                  , configMaxDiff         = defMaxDiff
                  , configCacheRefresh    = defCacheRefresh
                  , configCacheRetryDelay = defCacheRetryDelay
-                 , configNumTxId         = defNumTxId
                  , configStatsd          = defStatsd
                  , configStatsdHost      = defStatsdHost
                  , configStatsdPort      = defStatsdPort
                  , configStatsdPrefix    = defStatsdPrefix
+                 , configWebPriceGet     = defWebPriceGet
                  }
 
 defEnv :: MonadIO m => String -> a -> (String -> Maybe a) -> m a
@@ -190,11 +190,6 @@ defDebug = unsafePerformIO $
     defEnv "DEBUG" False parseBool
 {-# NOINLINE defDebug #-}
 
-defNumTxId :: Bool
-defNumTxId = unsafePerformIO $
-    defEnv "NUMTXID" False parseBool
-{-# NOINLINE defNumTxId #-}
-
 defWebLimits :: WebLimits
 defWebLimits = unsafePerformIO $ do
     max_limit  <- defEnv "MAX_LIMIT" (maxLimitCount def) readMaybe
@@ -271,6 +266,11 @@ defStatsdPort = unsafePerformIO $
     defEnv "STATSD_PORT" 8125 readMaybe
 {-# NOINLINE defStatsdPort #-}
 
+defWebPriceGet :: Int
+defWebPriceGet = unsafePerformIO $
+    defEnv "WEB_PRICE_GET" (90 * 1000 * 1000) readMaybe
+{-# NOINLINE defWebPriceGet #-}
+
 defStatsdPrefix :: String
 defStatsdPrefix =
     unsafePerformIO $
@@ -497,10 +497,6 @@ config = do
         <> help "Maximum difference between headers and blocks"
         <> showDefault
         <> value (configMaxDiff def)
-    configNumTxId <-
-        flag (configNumTxId def) True $
-        long "numtxid"
-        <> help "Numeric tx_index field"
     configStatsd <-
         flag (configStatsd def) True $
         long "statsd"
@@ -526,6 +522,13 @@ config = do
         <> help "Prefix for statsd metrics"
         <> showDefault
         <> value (configStatsdPrefix def)
+    configWebPriceGet <-
+        option auto $
+        metavar "MICROSECONDS"
+        <> long "web-price-get"
+        <> help "How often to retrieve price information"
+        <> showDefault
+        <> value (configWebPriceGet def)
     pure
         Config
             { configWebLimits = WebLimits {..}
@@ -598,11 +601,11 @@ run Config { configHost = host
            , configNoMempool = nomem
            , configCacheRefresh = crefresh
            , configCacheRetryDelay = cretrydelay
-           , configNumTxId = numtxid
            , configStatsd = statsd
            , configStatsdHost = statsdhost
            , configStatsdPort = statsdport
            , configStatsdPrefix = statsdpfx
+           , configWebPriceGet = wpget
            } =
     runStderrLoggingT . filterLogger l . with_stats $ \stats -> do
         $(logInfoS) "Main" $
@@ -645,8 +648,8 @@ run Config { configHost = host
                     , webVersion = version
                     , webMaxDiff = maxdiff
                     , webNoMempool = nomem
-                    , webNumTxId = numtxid
                     , webStats = stats
+                    , webPriceGet = wpget
                     }
   where
     with_stats go
