@@ -79,9 +79,8 @@ import           Haskoin.Store.Database.Reader
 import           Haskoin.Store.Database.Writer
 import           Haskoin.Store.Logic           (ImportException (Orphan),
                                                 deleteUnconfirmedTx,
-                                                getOldMempool, importBlock,
-                                                initBest, newMempoolTx,
-                                                revertBlock)
+                                                importBlock, initBest,
+                                                newMempoolTx, revertBlock)
 import           Haskoin.Store.Stats
 import           NQE                           (Listen, Mailbox, Publisher,
                                                 inboxToMailbox, newInbox,
@@ -171,7 +170,7 @@ setStoreHeight =
         Nothing -> setit m 0
         Just bb -> getBlock bb >>= \case
             Nothing -> setit m 0
-            Just b -> setit m (blockDataHeight b)
+            Just b  -> setit m (blockDataHeight b)
   where
     setit m i = liftIO $ storeHeight m `Metrics.Gauge.set` fromIntegral i
 
@@ -274,6 +273,8 @@ instance MonadIO m => StoreReadExtra (BlockT m) where
         runRocksDB . getAddressUnspents a
     getAddressTxs a =
         runRocksDB . getAddressTxs a
+    getNumTxData =
+        runRocksDB . getNumTxData
 
 -- | Run block store process.
 withBlockStore ::
@@ -326,7 +327,8 @@ withBlockStore cfg action = do
                       "Could not initialize: " <> cs (show e)
                   throwIO e
               Right () -> return ()
-    run inbox = withAsync (pingMe (inboxToMailbox inbox))
+    run inbox =
+          withAsync (pingMe (inboxToMailbox inbox))
           $ const
           $ forever
           $ receive inbox >>=
@@ -727,27 +729,6 @@ checkTime =
                     "Syncing peer timeout: " <> peerText p
                 killPeer PeerTimeout p
 
-pruneMempool :: MonadLoggerIO m => BlockT m ()
-pruneMempool = guardMempool $ do
-    s <- isInSync
-    when s $ do
-        now <- liftIO getCurrentTime
-        let seconds = floor (utcTimeToPOSIXSeconds now)
-        getOldMempool seconds >>= \old ->
-            unless (null old) $
-            runImport (mapM_ delete_it old) >>= \case
-                Left e -> do
-                    $(logErrorS) "BlockStore" $
-                        "Could not prune mempool: "
-                        <> cs (show e)
-                    throwIO e
-                Right x -> return x
-  where
-    delete_it txid = do
-        $(logDebugS) "BlockStore" $
-            "Deleting old mempool tx: " <> txHashToHex txid
-        deleteUnconfirmedTx False txid
-
 revertToMainChain :: MonadLoggerIO m => BlockT m ()
 revertToMainChain = do
     h <- headerHash . nodeHeader <$> getBest
@@ -962,26 +943,25 @@ processBlockStoreMessage (TxRefAvailable p ts) =
     processTxs p ts
 
 processBlockStoreMessage (BlockPing r) = do
-    trySyncing
-    processMempool
-    pruneOrphans
-    checkTime
-    pruneMempool
     setStoreHeight
     setHeadersHeight
     setPendingTxs
     setPeersConnected
     setMempoolSize
+    trySyncing
+    processMempool
+    pruneOrphans
+    checkTime
     atomically (r ())
 
 pingMe :: MonadLoggerIO m => Mailbox BlockStoreMessage -> m ()
 pingMe mbox =
     forever $ do
+        BlockPing `query` mbox
         delay <- liftIO $
             randomRIO (  100 * 1000
                       , 1000 * 1000 )
         threadDelay delay
-        BlockPing `query` mbox
 
 blockStorePeerConnect :: MonadIO m => Peer -> BlockStore -> m ()
 blockStorePeerConnect peer store =
