@@ -24,7 +24,6 @@ module Haskoin.Store.Cache
     , cacheNewBlock
     , cacheWriter
     , isInCache
-    , evictFromCache
     ) where
 
 import           Control.DeepSeq             (NFData)
@@ -35,11 +34,11 @@ import           Control.Monad.Logger        (MonadLoggerIO, logDebugS,
 import           Control.Monad.Reader        (ReaderT (..), ask, asks)
 import           Control.Monad.Trans         (lift)
 import           Control.Monad.Trans.Maybe   (MaybeT (..), runMaybeT)
-import           Data.Bits                   (shift, (.&.), (.|.))
+import           Data.Bits                   (complement, shift, (.&.), (.|.))
 import           Data.ByteString             (ByteString)
-import qualified Data.ByteString.Short       as BSS
+import qualified Data.ByteString             as B
 import           Data.Default                (def)
-import           Data.Either                 (rights)
+import           Data.Either                 (fromRight, isRight, rights)
 import           Data.HashMap.Strict         (HashMap)
 import qualified Data.HashMap.Strict         as HashMap
 import           Data.HashSet                (HashSet)
@@ -47,7 +46,8 @@ import qualified Data.HashSet                as HashSet
 import qualified Data.IntMap.Strict          as I
 import           Data.List                   (sort)
 import qualified Data.Map.Strict             as Map
-import           Data.Maybe                  (catMaybes, isNothing, mapMaybe)
+import           Data.Maybe                  (catMaybes, fromMaybe, isNothing,
+                                              mapMaybe)
 import           Data.Serialize              (Serialize, decode, encode)
 import           Data.String.Conversions     (cs)
 import           Data.Text                   (Text)
@@ -94,8 +94,9 @@ import qualified System.Metrics.Gauge        as Metrics.Gauge
 import           System.Random               (randomIO, randomRIO)
 import           UnliftIO                    (Exception, MonadIO, MonadUnliftIO,
                                               TQueue, TVar, async, atomically,
-                                              bracket, liftIO, link, readTQueue,
-                                              readTVar, throwIO, withAsync,
+                                              bracket, liftIO, link, modifyTVar,
+                                              newTVarIO, readTQueue, readTVar,
+                                              throwIO, wait, withAsync,
                                               writeTQueue, writeTVar)
 import           UnliftIO.Concurrent         (threadDelay)
 
@@ -119,11 +120,16 @@ data CacheConfig = CacheConfig
     }
 
 data CacheMetrics = CacheMetrics
-    { cacheHits      :: !Metrics.Counter
-    , cacheMisses    :: !Metrics.Counter
-    , cacheIgnore    :: !Metrics.Counter
-    , cacheRefreshes :: !Metrics.Counter
-    , cacheIndexTime :: !StatDist
+    { cacheHits            :: !Metrics.Counter
+    , cacheMisses          :: !Metrics.Counter
+    , cacheIgnore          :: !Metrics.Counter
+    , cacheRefreshes       :: !Metrics.Counter
+    , cacheLockAcquired    :: !Metrics.Counter
+    , cacheLockReleased    :: !Metrics.Counter
+    , cacheLockFailed      :: !Metrics.Counter
+    , cacheIndexTime       :: !StatDist
+    , cacheMempoolSyncTime :: !StatDist
+    , cacheBlockSyncTime   :: !StatDist
     }
 
 newCacheMetrics :: MonadIO m => Metrics.Store -> m CacheMetrics
@@ -132,7 +138,12 @@ newCacheMetrics s = liftIO $ do
     cacheMisses          <- c "misses"
     cacheIgnore          <- c "ignore"
     cacheRefreshes       <- c "refreshes"
-    cacheIndexTime       <- d "index_time_ms"
+    cacheLockAcquired    <- c "lock_acquired"
+    cacheLockReleased    <- c "lock_released"
+    cacheLockFailed      <- c "lock_failed"
+    cacheIndexTime       <- d "index"
+    cacheMempoolSyncTime <- d "mempool_sync"
+    cacheBlockSyncTime   <- d "block_sync"
     return CacheMetrics{..}
   where
     c x = Metrics.createCounter ("cache." <> x) s
@@ -154,7 +165,9 @@ withMetrics df go =
     end metrics t1 = do
         t2 <- systemToUTCTime <$> liftIO getSystemTime
         let diff = round $ diffUTCTime t2 t1 * 1000
-        df metrics `addStatEntry` StatEntry diff 1
+        df metrics `addStatTime` diff
+        df metrics `addStatItems` 1
+        addStatQuery (df metrics)
 
 incrementCounter :: MonadIO m
                  => (CacheMetrics -> Metrics.Counter)
@@ -201,20 +214,33 @@ instance (MonadUnliftIO m , MonadLoggerIO m, StoreReadExtra m) =>
     getAddressTxs addr = lift . getAddressTxs addr
     getAddressUnspents addr = lift . getAddressUnspents addr
     getAddressesUnspents addrs = lift . getAddressesUnspents addrs
-    xPubBals xpub =
-        ask >>= \case
-            Nothing  -> lift (xPubBals xpub)
-            Just cfg -> lift (runReaderT (getXPubBalances xpub) cfg)
-    xPubUnspents xpub limits =
-        ask >>= \case
-            Nothing  -> lift (xPubUnspents xpub limits)
-            Just cfg -> lift (runReaderT (getXPubUnspents xpub limits) cfg)
-    xPubTxs xpub limits =
-        ask >>= \case
-            Nothing  -> lift (xPubTxs xpub limits)
-            Just cfg -> lift (runReaderT (getXPubTxs xpub limits) cfg)
     getMaxGap = lift getMaxGap
     getInitialGap = lift getInitialGap
+    getNumTxData = lift . getNumTxData
+    xPubBals xpub =
+        ask >>= \case
+            Nothing  -> lift $
+                xPubBals xpub
+            Just cfg -> lift $
+                runReaderT (getXPubBalances xpub) cfg
+    xPubUnspents xpub xbals limits =
+        ask >>= \case
+            Nothing  -> lift $
+                xPubUnspents xpub xbals limits
+            Just cfg -> lift $
+                runReaderT (getXPubUnspents xpub xbals limits) cfg
+    xPubTxs xpub xbals limits =
+        ask >>= \case
+            Nothing  -> lift $
+                xPubTxs xpub xbals limits
+            Just cfg -> lift $
+                runReaderT (getXPubTxs xpub xbals limits) cfg
+    xPubTxCount xpub xbals =
+        ask >>= \case
+            Nothing  -> lift $
+                xPubTxCount xpub xbals
+            Just cfg -> lift $
+                runReaderT (getXPubTxCount xpub xbals) cfg
 
 withCache :: StoreReadBase m => Maybe CacheConfig -> CacheT m a -> m a
 withCache s f = runReaderT f s
@@ -228,77 +254,93 @@ txSetPfx = "t"
 utxoPfx :: ByteString
 utxoPfx = "u"
 
+idxPfx :: ByteString
+idxPfx = "i"
+
 getXPubTxs ::
        (MonadUnliftIO m, MonadLoggerIO m, StoreReadExtra m)
-    => XPubSpec -> Limits -> CacheX m [TxRef]
-getXPubTxs xpub limits = do
-    xpubtxt <- xpubText xpub
-    isXPubCached xpub >>= \case
-        True -> do
-            incrementCounter cacheHits
-            txs <- cacheGetXPubTxs xpub limits
-            return txs
-        False -> do
-            bals <- lift $ xPubBals xpub
-            newXPubC xpub bals
-            lift $ xPubBalsTxs bals limits
-
-getXPubUnspents ::
-       (MonadUnliftIO m, MonadLoggerIO m, StoreReadExtra m)
-    => XPubSpec -> Limits -> CacheX m [XPubUnspent]
-getXPubUnspents xpub limits = do
-    xpubtxt <- xpubText xpub
-    isXPubCached xpub >>= \case
-        True -> do
-            incrementCounter cacheHits
-            bals <- cacheGetXPubBalances xpub
-            process bals
-        False -> do
-            bals <- lift $ xPubBals xpub
-            newXPubC xpub bals
-            lift $ xPubBalsUnspents bals limits
+    => XPubSpec -> [XPubBal] -> Limits -> CacheX m [TxRef]
+getXPubTxs xpub xbals limits = go False
   where
-    process bals = do
+    go m = isXPubCached xpub >>= \case
+        True -> do
+            unless m $ incrementCounter cacheHits
+            cacheGetXPubTxs xpub limits
+        False ->
+            case m of
+                True -> lift $ xPubTxs xpub xbals limits
+                False -> do
+                    newXPubC xpub xbals
+                    go True
+
+getXPubTxCount :: (MonadUnliftIO m, MonadLoggerIO m, StoreReadExtra m)
+               => XPubSpec -> [XPubBal] -> CacheX m Word32
+getXPubTxCount xpub xbals =
+    go False
+  where
+    go t = isXPubCached xpub >>= \case
+        True -> do
+            unless t $ incrementCounter cacheHits
+            cacheGetXPubTxCount xpub
+        False ->
+            if t
+            then lift $ xPubTxCount xpub xbals
+            else do
+                newXPubC xpub xbals
+                go True
+
+getXPubUnspents :: (MonadUnliftIO m, MonadLoggerIO m, StoreReadExtra m)
+                => XPubSpec -> [XPubBal] -> Limits -> CacheX m [XPubUnspent]
+getXPubUnspents xpub xbals limits =
+    go False
+  where
+    xm = let f x = (balanceAddress (xPubBal x), x)
+             g = (> 0) . balanceUnspentCount . xPubBal
+         in HashMap.fromList $ map f $ filter g xbals
+    go m = isXPubCached xpub >>= \case
+        True -> do
+            unless m $ incrementCounter cacheHits
+            process
+        False -> case m of
+            True ->
+                lift $ xPubUnspents xpub xbals limits
+            False -> do
+                newXPubC xpub xbals
+                go True
+    process = do
         ops <- map snd <$> cacheGetXPubUnspents xpub limits
         uns <- catMaybes <$> lift (mapM getUnspent ops)
-        let addrmap =
-                Map.fromList $
-                map (\b -> (balanceAddress (xPubBal b), xPubBalPath b)) bals
-            addrutxo =
-                mapMaybe
-                    (\u ->
-                         either
-                             (const Nothing)
-                             (\a -> Just (a, u))
-                             (scriptToAddressBS
-                                  (BSS.fromShort (unspentScript u))))
-                    uns
-            xpubutxo =
-                mapMaybe
-                    (\(a, u) -> (`XPubUnspent` u) <$> Map.lookup a addrmap)
-                    addrutxo
-        xpubtxt <- xpubText xpub
-        return xpubutxo
+        let f u = either
+                  (const Nothing)
+                  (\a -> Just (a, u))
+                  (scriptToAddressBS (unspentScript u))
+            g a = HashMap.lookup a xm
+            h u x =
+                XPubUnspent
+                {
+                    xPubUnspent = u,
+                    xPubUnspentPath = xPubBalPath x
+                }
+            us = mapMaybe f uns
+            i a u = h u <$> g a
+        return $ mapMaybe (uncurry i) us
 
-getXPubBalances ::
-       (MonadUnliftIO m, MonadLoggerIO m, StoreReadExtra m)
-    => XPubSpec -> CacheX m [XPubBal]
-getXPubBalances xpub = do
-    xpubtxt <- xpubText xpub
-    isXPubCached xpub >>= \case
-        True -> do
-            incrementCounter cacheHits
-            cacheGetXPubBalances xpub
-        False -> do
-            bals <- lift $ xPubBals xpub
-            newXPubC xpub bals
-            return bals
+getXPubBalances :: (MonadUnliftIO m, MonadLoggerIO m, StoreReadExtra m)
+                => XPubSpec -> CacheX m [XPubBal]
+getXPubBalances xpub = isXPubCached xpub >>= \case
+    True -> do
+        incrementCounter cacheHits
+        cacheGetXPubBalances xpub
+    False -> do
+        bals <- lift $ xPubBals xpub
+        newXPubC xpub bals
+        return bals
 
 isInCache :: MonadLoggerIO m => XPubSpec -> CacheT m Bool
 isInCache xpub =
     ask >>= \case
-        Nothing  -> return False
-        Just cfg -> runReaderT (isXPubCached xpub) cfg
+    Nothing  -> return False
+    Just cfg -> runReaderT (isXPubCached xpub) cfg
 
 isXPubCached :: MonadLoggerIO m => XPubSpec -> CacheX m Bool
 isXPubCached = runRedis . redisIsXPubCached
@@ -312,32 +354,60 @@ cacheGetXPubBalances xpub = do
     touchKeys [xpub]
     return bals
 
+cacheGetXPubTxCount :: (MonadUnliftIO m, MonadLoggerIO m, StoreReadExtra m)
+                    => XPubSpec -> CacheX m Word32
+cacheGetXPubTxCount xpub = do
+    count <- fromInteger <$> runRedis (redisGetXPubTxCount xpub)
+    touchKeys [xpub]
+    return count
+
+redisGetXPubTxCount :: RedisCtx m f => XPubSpec -> m (f Integer)
+redisGetXPubTxCount xpub = Redis.zcard (txSetPfx <> encode xpub)
+
 cacheGetXPubTxs ::
        (StoreReadBase m, MonadLoggerIO m)
     => XPubSpec
     -> Limits
     -> CacheX m [TxRef]
-cacheGetXPubTxs xpub limits = do
-    score <-
-        case start limits of
-            Nothing -> return Nothing
-            Just (AtTx th) ->
-                lift (getTxData th) >>= \case
-                    Just TxData {txDataBlock = b@BlockRef {}} ->
-                        return (Just (blockRefScore b))
-                    _ -> return Nothing
-            Just (AtBlock h) ->
-                return (Just (blockRefScore (BlockRef h maxBound)))
-    xs <-
-        runRedis $
-        getFromSortedSet
-            (txSetPfx <> encode xpub)
-            score
-            (offset limits)
-            (limit limits)
-    touchKeys [xpub]
-    return $ map (uncurry f) xs
+cacheGetXPubTxs xpub limits =
+    case start limits of
+        Nothing ->
+            go1 Nothing
+        Just (AtTx th) -> lift (getTxData th) >>= \case
+            Just TxData {txDataBlock = b@BlockRef{}} ->
+                go1 $ Just (blockRefScore b)
+            _ ->
+                go2 th
+        Just (AtBlock h) ->
+            go1 (Just (blockRefScore (BlockRef h maxBound)))
   where
+    go1 score = do
+        xs <- runRedis $
+            getFromSortedSet
+                (txSetPfx <> encode xpub)
+                score
+                (offset limits)
+                (limit limits)
+        touchKeys [xpub]
+        return $ map (uncurry f) xs
+    go2 hash = do
+        xs <- runRedis $
+            getFromSortedSet
+                (txSetPfx <> encode xpub)
+                Nothing
+                0
+                0
+        touchKeys [xpub]
+        let xs' = if any ((== hash) . fst) xs
+                  then dropWhile ((/= hash) . fst) xs
+                  else []
+        return
+            $ map (uncurry f)
+            $ l
+            $ drop (fromIntegral (offset limits)) xs'
+    l = if limit limits > 0
+        then take (fromIntegral (limit limits))
+        else id
     f t s = TxRef {txRefHash = t, txRefBlock = scoreBlockRef s}
 
 cacheGetXPubUnspents ::
@@ -345,27 +415,46 @@ cacheGetXPubUnspents ::
     => XPubSpec
     -> Limits
     -> CacheX m [(BlockRef, OutPoint)]
-cacheGetXPubUnspents xpub limits = do
-    score <-
-        case start limits of
-            Nothing -> return Nothing
-            Just (AtTx th) ->
-                lift (getTxData th) >>= \case
-                    Just TxData {txDataBlock = b@BlockRef {}} ->
-                        return (Just (blockRefScore b))
-                    _ -> return Nothing
-            Just (AtBlock h) ->
-                return (Just (blockRefScore (BlockRef h maxBound)))
-    xs <-
-        runRedis $
-        getFromSortedSet
-            (utxoPfx <> encode xpub)
-            score
-            (offset limits)
-            (limit limits)
-    touchKeys [xpub]
-    return $ map (uncurry f) xs
+cacheGetXPubUnspents xpub limits =
+    case start limits of
+        Nothing ->
+            go1 Nothing
+        Just (AtTx th) -> lift (getTxData th) >>= \case
+            Just TxData {txDataBlock = b@BlockRef{}} ->
+                go1 (Just (blockRefScore b))
+            _ ->
+                go2 th
+        Just (AtBlock h) ->
+            go1 (Just (blockRefScore (BlockRef h maxBound)))
   where
+    go1 score = do
+        xs <-
+            runRedis $
+            getFromSortedSet
+                (utxoPfx <> encode xpub)
+                score
+                (offset limits)
+                (limit limits)
+        touchKeys [xpub]
+        return $ map (uncurry f) xs
+    go2 hash = do
+        xs <- runRedis $
+            getFromSortedSet
+                (utxoPfx <> encode xpub)
+                Nothing
+                0
+                0
+        touchKeys [xpub]
+        let xs' = if any ((== hash) . outPointHash . fst) xs
+                  then dropWhile ((/= hash) . outPointHash . fst) xs
+                  else []
+        return
+            $ map (uncurry f)
+            $ l
+            $ drop (fromIntegral (offset limits)) xs'
+    l = if limit limits > 0
+        then take (fromIntegral (limit limits))
+        else id
     f o s = (scoreBlockRef s, o)
 
 redisGetXPubBalances :: (Functor f, RedisCtx m f) => XPubSpec -> m (f [XPubBal])
@@ -416,7 +505,13 @@ getFromSortedSet key Nothing off count = do
         ys <- map (\(x, s) -> (, s) <$> decode x) <$> xs
         return (rights ys)
 getFromSortedSet key (Just score) off 0 = do
-    xs <- zrangebyscoreWithscoresLimit key score (1 / 0) (fromIntegral off) (-1)
+    xs <-
+        zrangebyscoreWithscoresLimit
+            key
+            score
+            (1 / 0)
+            (fromIntegral off)
+            (-1)
     return $ do
         ys <- map (\(x, s) -> (, s) <$> decode x) <$> xs
         return (rights ys)
@@ -500,21 +595,26 @@ lockIt = do
         Right Redis.Ok -> do
             $(logDebugS) "Cache" $
                 "Acquired lock with value " <> cs (show rnd)
+            incrementCounter cacheLockAcquired
             return (Just rnd)
         Right Redis.Pong -> do
             $(logErrorS) "Cache"
                 "Unexpected pong when acquiring lock"
+            incrementCounter cacheLockFailed
             return Nothing
         Right (Redis.Status s) -> do
             $(logErrorS) "Cache" $
                 "Unexpected status acquiring lock: " <> cs s
+            incrementCounter cacheLockFailed
             return Nothing
         Left (Redis.Bulk Nothing) -> do
             $(logDebugS) "Cache" "Lock already taken"
+            incrementCounter cacheLockFailed
             return Nothing
         Left e -> do
             $(logErrorS) "Cache"
                 "Error when trying to acquire lock"
+            incrementCounter cacheLockFailed
             throwIO (RedisError e)
   where
     go rnd = do
@@ -544,6 +644,7 @@ unlockIt (Just i) =
                 $(logDebugS) "Cache" $
                     "Released lock with value " <>
                     cs (show i)
+                incrementCounter cacheLockReleased
             else
                 $(logErrorS) "Cache" $
                     "Could not release lock: value is not " <>
@@ -603,7 +704,7 @@ pruneDB = do
     flush n =
         case n `div` 64 of
         0 -> return 0
-        x -> withLockForever $ do
+        x -> fmap (fromMaybe 0) $ withLock $ do
             ks <- fmap (map fst) . runRedis $
                   getFromSortedSet maxKey Nothing 0 (fromIntegral x)
             $(logDebugS) "Cache" $
@@ -624,22 +725,25 @@ cacheWriterReact ::
        (MonadUnliftIO m, MonadLoggerIO m, StoreReadExtra m)
     => CacheWriterMessage -> CacheX m ()
 cacheWriterReact CacheNewBlock =
-    inSync >>= \s -> when s newBlockC
-cacheWriterReact (CachePing respond) = do
-    s <- inSync
-    when s $ do
+    doSync
+cacheWriterReact (CachePing respond) =
+    doSync >> atomically (respond ())
+
+doSync :: (MonadUnliftIO m, MonadLoggerIO m, StoreReadExtra m)
+       => CacheX m ()
+doSync =
+    inSync >>= \s -> when s $ do
+        newBlockC
         syncMempoolC
         void pruneDB
-    atomically (respond ())
 
 lenNotNull :: [XPubBal] -> Int
-lenNotNull bals = length $ filter (not . nullBalance . xPubBal) bals
+lenNotNull = length . filter (not . nullBalance . xPubBal)
 
 newXPubC ::
        (MonadUnliftIO m, MonadLoggerIO m, StoreReadExtra m)
     => XPubSpec -> [XPubBal] -> CacheX m ()
-newXPubC xpub bals =
-    should_index >>= \i ->
+newXPubC xpub xbals = should_index >>= \i ->
     if i
     then do
         incrementCounter cacheMisses
@@ -647,33 +751,44 @@ newXPubC xpub bals =
     else incrementCounter cacheIgnore
   where
     op XPubUnspent {xPubUnspent = u} = (unspentPoint u, unspentBlock u)
-    should_index = do
-        x <- asks cacheMin
-        if x <= lenNotNull bals
-            then inSync
-            else return False
-    index = do
+    should_index =
+        asks cacheMin >>= \x ->
+        if x <= lenNotNull xbals then inSync else return False
+    index =
+        bracket set_index unset_index $ \y -> when y $
+        withMetrics cacheIndexTime $ do
         xpubtxt <- xpubText xpub
         $(logDebugS) "Cache" $
-            "Caching " <> xpubtxt <> ": " <> cs (show (length bals)) <>
-            " addresses / " <> cs (show (lenNotNull bals)) <>
-            " used"
-        utxo <- lift $ xPubUnspents xpub def
+            "Caching " <> xpubtxt <> ": " <>
+            cs (show (length xbals)) <> " addresses / " <>
+            cs (show (lenNotNull xbals)) <> " used"
+        utxo <- lift $ xPubUnspents xpub xbals def
         $(logDebugS) "Cache" $
             "Caching " <> xpubtxt <> ": " <> cs (show (length utxo)) <>
             " utxos"
-        xtxs <- lift $ xPubTxs xpub def
+        xtxs <- lift $ xPubTxs xpub xbals def
         $(logDebugS) "Cache" $
             "Caching " <> xpubtxt <> ": " <> cs (show (length xtxs)) <>
             " txs"
         now <- systemSeconds <$> liftIO getSystemTime
         runRedis $ do
             b <- redisTouchKeys now [xpub]
-            c <- redisAddXPubBalances xpub bals
+            c <- redisAddXPubBalances xpub xbals
             d <- redisAddXPubUnspents xpub (map op utxo)
             e <- redisAddXPubTxs xpub xtxs
             return $ b >> c >> d >> e >> return ()
         $(logDebugS) "Cache" $ "Cached " <> xpubtxt
+    key = idxPfx <> encode xpub
+    opts = Redis.SetOpts
+           { Redis.setSeconds = Just 600
+           , Redis.setMilliseconds = Nothing
+           , Redis.setCondition = Just Redis.Nx
+           }
+    red = Redis.setOpts key "1" opts
+    unset_index y = when y . void . runRedis $ Redis.del [key]
+    set_index =
+        asks cacheConn >>= \conn ->
+        liftIO (Redis.runRedis conn red) >>= return . isRight
 
 inSync :: (MonadUnliftIO m, MonadLoggerIO m, StoreReadExtra m)
        => CacheX m Bool
@@ -697,7 +812,7 @@ newBlockC =
             cacheGetHead >>= \case
                 Nothing -> do
                     $(logInfoS) "Cache" "Initializing best cache block"
-                    importBlockC bb
+                    void $ importBlockC bb
                 Just hb ->
                     if hb == headerHash (nodeHeader cn)
                     then $(logDebugS) "Cache" "Cache in sync"
@@ -727,11 +842,10 @@ newBlockC =
                         $(logDebugS) "Cache" $
                             "Reverting cache head not in main chain: " <>
                             blockHashToHex hb
-                        removeHeadC hb
-                        newBlockC
+                        removeHeadC hb >>= \s -> when s newBlockC
     next ch bn hn =
         if | prevBlock (nodeHeader bn) == headerHash (nodeHeader hn) ->
-                 importBlockC (headerHash (nodeHeader bn))
+                 void $ importBlockC (headerHash (nodeHeader bn))
            | nodeHeight bn > nodeHeight hn ->
                  chainGetAncestor (nodeHeight hn + 1) bn ch >>= \case
                      Nothing ->
@@ -740,16 +854,17 @@ newBlockC =
                          <> cs (show (nodeHeight hn + 1))
                          <> " for block: "
                          <> blockHashToHex (headerHash (nodeHeader bn))
-                     Just hn' -> do
-                         importBlockC (headerHash (nodeHeader hn'))
-                         newBlockC
+                     Just hn' ->
+                         importBlockC (headerHash (nodeHeader hn')) >>= \s ->
+                         when s newBlockC
            | otherwise ->
                  $(logInfoS) "Cache" "Cache best block higher than this node's"
 
 importBlockC :: (MonadUnliftIO m, StoreReadExtra m, MonadLoggerIO m)
-             => BlockHash -> CacheX m ()
+             => BlockHash -> CacheX m Bool
 importBlockC bh =
-    withLockForever $
+    fmap (maybe False (const True)) $ withLock $
+    withMetrics cacheBlockSyncTime $
     cacheGetHead >>= \case
         Nothing -> go
         Just cb ->
@@ -784,9 +899,9 @@ importBlockC bh =
                 <> blockHashToHex bh
 
 removeHeadC :: (StoreReadExtra m, MonadUnliftIO m, MonadLoggerIO m)
-            => BlockHash -> CacheX m ()
+            => BlockHash -> CacheX m Bool
 removeHeadC cb =
-    withLockForever $
+    fmap (maybe False (const True)) $ withLock $
     void . runMaybeT $ do
     bh <- MaybeT cacheGetHead
     guard (cb == bh)
@@ -862,7 +977,7 @@ importMultiTxC txs = do
     getxbals xpubs = do
         bals <- runRedis . fmap sequence . forM xpubs $ \xpub -> do
             bs <- redisGetXPubBalances xpub
-            return $ (xpub, ) <$> bs
+            return $ (,) xpub <$> bs
         return $ HashMap.filter (not . null) (HashMap.fromList bals)
     allops = map snd $ concatMap txInputs txs <> concatMap txOutputs txs
     alladdrs =
@@ -1035,7 +1150,8 @@ syncMempoolC :: (MonadUnliftIO m, MonadLoggerIO m, StoreReadExtra m)
              => CacheX m ()
 syncMempoolC =
     toInteger <$> asks cacheRefresh >>= \refresh ->
-    void . withLockForever . withCool "cool" (refresh * 9 `div` 10) $ do
+    void . withLock . withCool "cool" (refresh * 9 `div` 10) $
+    withMetrics cacheMempoolSyncTime $ do
     nodepool <- HashSet.fromList . map snd <$> lift getMempool
     cachepool <- HashSet.fromList . map snd <$> cacheGetMempool
     getem (HashSet.difference nodepool cachepool)
@@ -1082,14 +1198,6 @@ redisSetAddrInfo ::
        (Functor f, RedisCtx m f) => Address -> AddressXPub -> m (f ())
 redisSetAddrInfo a i = void <$> Redis.set (addrPfx <> encode a) (encode i)
 
-evictFromCache ::
-       (MonadUnliftIO m, MonadLoggerIO m, StoreReadBase m)
-    => [XPubSpec]
-    -> CacheT m ()
-evictFromCache xpubs = ask >>= \case
-    Nothing  -> return ()
-    Just cfg -> void (runReaderT (withLockRetry 100 (delXPubKeys xpubs)) cfg)
-
 delXPubKeys ::
        (MonadUnliftIO m, MonadLoggerIO m, StoreReadBase m)
     => [XPubSpec]
@@ -1143,27 +1251,32 @@ redisAddXPubUnspents ::
     => XPubSpec
     -> [(OutPoint, BlockRef)]
     -> m (f Integer)
-redisAddXPubUnspents _ [] = return (pure 0)
+redisAddXPubUnspents _ [] =
+    return (pure 0)
 redisAddXPubUnspents xpub utxo =
     zadd (utxoPfx <> encode xpub) $
     map (\(p, r) -> (blockRefScore r, encode p)) utxo
 
 redisRemXPubUnspents ::
        (Applicative f, RedisCtx m f) => XPubSpec -> [OutPoint] -> m (f Integer)
-redisRemXPubUnspents _ []     = return (pure 0)
-redisRemXPubUnspents xpub ops = zrem (utxoPfx <> encode xpub) (map encode ops)
+redisRemXPubUnspents _ [] =
+    return (pure 0)
+redisRemXPubUnspents xpub ops =
+    zrem (utxoPfx <> encode xpub) (map encode ops)
 
 redisAddXPubBalances ::
        (Monad f, RedisCtx m f) => XPubSpec -> [XPubBal] -> m (f ())
 redisAddXPubBalances _ [] = return (pure ())
 redisAddXPubBalances xpub bals = do
     xs <- mapM (uncurry (Redis.hset (balancesPfx <> encode xpub))) entries
-    ys <-
-        forM bals $ \b ->
-            redisSetAddrInfo
-                (balanceAddress (xPubBal b))
-                AddressXPub
-                    {addressXPubSpec = xpub, addressXPubPath = xPubBalPath b}
+    ys <- forM bals $ \b ->
+        redisSetAddrInfo
+        (balanceAddress (xPubBal b))
+        AddressXPub
+        {
+            addressXPubSpec = xpub,
+            addressXPubPath = xPubBalPath b
+        }
     return $ sequence_ xs >> sequence_ ys
   where
     entries = map (\b -> (encode (xPubBalPath b), encode (xPubBal b))) bals
@@ -1246,7 +1359,11 @@ xpubText :: (MonadUnliftIO m, MonadLoggerIO m, StoreReadBase m)
          => XPubSpec -> CacheX m Text
 xpubText xpub = do
     net <- lift getNetwork
-    return . cs $ xPubExport net (xPubSpecKey xpub)
+    let suffix = case xPubDeriveType xpub of
+                     DeriveNormal -> ""
+                     DeriveP2SH   -> "/p2sh"
+                     DeriveP2WPKH -> "/p2wpkh"
+    return . cs $ suffix <> xPubExport net (xPubSpecKey xpub)
 
 cacheNewBlock :: MonadIO m => CacheWriter -> m ()
 cacheNewBlock = send CacheNewBlock
