@@ -9,41 +9,42 @@ module Haskoin.Store.Logic
     ( ImportException (..)
     , MonadImport
     , initBest
-    , getOldMempool
     , revertBlock
     , importBlock
     , newMempoolTx
     , deleteUnconfirmedTx
     ) where
 
-import           Control.Monad         (forM, forM_, guard, unless, void, when,
-                                        zipWithM_)
-import           Control.Monad.Except  (MonadError, throwError)
-import           Control.Monad.Logger  (MonadLoggerIO (..), logDebugS,
-                                        logErrorS)
-import qualified Data.ByteString       as B
-import qualified Data.ByteString.Short as B.Short
-import           Data.Either           (rights)
-import qualified Data.IntMap.Strict    as I
-import           Data.List             (nub)
-import           Data.Maybe            (catMaybes, fromMaybe, isJust, isNothing)
-import           Data.Serialize        (encode)
-import           Data.Word             (Word32, Word64)
-import           Haskoin               (Address, Block (..), BlockHash,
-                                        BlockHeader (..), BlockNode (..),
-                                        Network (..), OutPoint (..), Tx (..),
-                                        TxHash, TxIn (..), TxOut (..),
-                                        blockHashToHex, computeSubsidy,
-                                        eitherToMaybe, genesisBlock,
-                                        genesisNode, headerHash, isGenesis,
-                                        nullOutPoint, scriptToAddressBS, txHash,
-                                        txHashToHex)
+import           Conduit              (ConduitT, await, lift, sealConduitT,
+                                       yield, ($$++))
+import           Control.Monad        (forM, forM_, guard, unless, void, when,
+                                       zipWithM_)
+import           Control.Monad.Except (MonadError, throwError)
+import           Control.Monad.Logger (MonadLoggerIO (..), logDebugS, logErrorS)
+import qualified Data.ByteString      as B
+import           Data.Either          (rights)
+import           Data.Function        (on)
+import qualified Data.IntMap.Strict   as I
+import           Data.List            (nub, sortBy)
+import qualified Data.Map.Strict      as Map
+import           Data.Maybe           (catMaybes, fromMaybe, isJust, isNothing,
+                                       mapMaybe)
+import           Data.Serialize       (encode)
+import           Data.Word            (Word32, Word64)
+import           Haskoin              (Address, Block (..), BlockHash,
+                                       BlockHeader (..), BlockNode (..),
+                                       Network (..), OutPoint (..), Tx (..),
+                                       TxHash, TxIn (..), TxOut (..),
+                                       blockHashToHex, computeSubsidy,
+                                       eitherToMaybe, genesisBlock, genesisNode,
+                                       headerHash, isGenesis, nullOutPoint,
+                                       scriptToAddressBS, txHash, txHashToHex)
 import           Haskoin.Store.Common
-import           Haskoin.Store.Data    (Balance (..), BlockData (..),
-                                        BlockRef (..), Prev (..), Spender (..),
-                                        TxData (..), TxRef (..), UnixTime,
-                                        Unspent (..), confirmed)
-import           UnliftIO              (Exception)
+import           Haskoin.Store.Data   (Balance (..), BlockData (..),
+                                       BlockRef (..), Prev (..), Spender (..),
+                                       TxData (..), TxRef (..), UnixTime,
+                                       Unspent (..), confirmed)
+import           UnliftIO             (Exception)
 
 type MonadImport m =
     ( MonadError ImportException m
@@ -163,7 +164,7 @@ importOrConfirm bn txns = do
     action i tx =
         testPresent tx >>= \case
             False -> import_it i tx
-            True -> confirm_it i tx
+            True  -> confirm_it i tx
     confirm_it i tx =
         getActiveTxData (txHash tx) >>= \case
             Just t -> do
@@ -272,7 +273,7 @@ prepareTxData rbf br tt tx us =
            , txDataTime = tt
            }
   where
-    mkprv u = Prev (B.Short.fromShort (unspentScript u)) (unspentAmount u)
+    mkprv u = Prev (unspentScript u) (unspentAmount u)
     ps = I.fromList $ zip [0 ..] $ map mkprv us
 
 importTx
@@ -330,7 +331,7 @@ adjustAddressOutput op o old new = do
                 { unspentBlock = new
                 , unspentPoint = op
                 , unspentAmount = outValue o
-                , unspentScript = B.Short.toShort pk
+                , unspentScript = pk
                 , unspentAddress = ma
                 }
         forM_ ma $ replace_addr_unspent pk
@@ -341,7 +342,7 @@ adjustAddressOutput op o old new = do
                 { unspentBlock = old
                 , unspentPoint = op
                 , unspentAmount = outValue o
-                , unspentScript = B.Short.toShort pk
+                , unspentScript = pk
                 , unspentAddress = Just a
                 }
         insertAddrUnspent
@@ -350,7 +351,7 @@ adjustAddressOutput op o old new = do
                 { unspentBlock = new
                 , unspentPoint = op
                 , unspentAmount = outValue o
-                , unspentScript = B.Short.toShort pk
+                , unspentScript = pk
                 , unspentAddress = Just a
                 }
         decreaseBalance (confirmed old) a (outValue o)
@@ -545,7 +546,7 @@ modOutput add br op o = do
     v | add = (+ outValue o)
       | otherwise = subtract (outValue o)
     ma = eitherToMaybe (scriptToAddressBS (scriptOutput o))
-    u = Unspent { unspentScript = B.Short.toShort (scriptOutput o)
+    u = Unspent { unspentScript = scriptOutput o
                 , unspentBlock = br
                 , unspentPoint = op
                 , unspentAmount = outValue o
@@ -578,11 +579,11 @@ getTxOut i tx = do
 spendOutput :: MonadImport m => TxHash -> Word32 -> OutPoint -> m ()
 spendOutput th ix op = do
     u <- getUnspent op >>= \case
-        Just u -> return u
+        Just u  -> return u
         Nothing -> error $ "Could not find UTXO to spend: " <> show op
     deleteUnspent op
     insertSpender op (Spender th ix)
-    let pk = B.Short.fromShort (unspentScript u)
+    let pk = unspentScript u
     forM_ (scriptToAddressBS pk) $ \a -> do
         decreaseBalance
             (confirmed (unspentBlock u))
@@ -597,14 +598,14 @@ unspendOutput :: MonadImport m => OutPoint -> m ()
 unspendOutput op = do
     t <- getActiveTxData (outPointHash op) >>= \case
         Nothing -> error $ "Could not find tx data: " <> show (outPointHash op)
-        Just t -> return t
+        Just t  -> return t
     let o = fromMaybe
             (error ("Could not find output: " <> show op))
             (getTxOut (outPointIndex op) (txData t))
         m = eitherToMaybe (scriptToAddressBS (scriptOutput o))
         u = Unspent { unspentAmount = outValue o
                     , unspentBlock = txDataBlock t
-                    , unspentScript = B.Short.toShort (scriptOutput o)
+                    , unspentScript = scriptOutput o
                     , unspentPoint = op
                     , unspentAddress = m
                     }
